# ======================================================================
# Copyright 2018 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import math
from distutils.util import strtobool
from typing import Optional, Tuple
from urllib.parse import quote

from supvisors.process import ProcessStatus
from supvisors.ttypes import StartingStrategies, NameList
from supvisors.utils import get_bit, set_bit
from .webutils import SUPVISORS_PAGE, error_message

# form parameters
SERVER_URL = 'SERVER_URL'
SERVER_PORT = 'SERVER_PORT'
PATH_TRANSLATED = 'PATH_TRANSLATED'

IDENTIFIER = 'ident'  # navigation
APPLI = 'appname'  # navigation

ACTION = 'action'
NAMESPEC = 'namespec'   # used for actions
PROCESS = 'processname'  # used to tail (but also to display statistics)

PERIOD = 'period'
STRATEGY = 'strategy'
CPU = 'cpuid'
NIC = 'nic'
DISK_STATS = 'diskstats'
PARTITION = 'partition'
DEVICE = 'device'
AUTO = 'auto'  # auto-refresh
LIMIT = 'limit'

MESSAGE = 'message'
GRAVITY = 'gravity'

APP_SHRINK_EXPAND = 'ashex'
PROC_SHRINK_EXPAND = 'pshex'

# response parameters
HEADERS = 'headers'
LOCATION = 'Location'


class ViewContext:
    """ Class used to retrieve the parameters selected on the web page.
    It is also used to format href in html pages. """

    def __init__(self, context):
        """ Define attributes for statistics selection. """
        # store the HTML context
        self.http_context = context
        # keep references to Supvisors instance
        self.supvisors = context.supervisord.supvisors
        self.logger = self.supvisors.logger
        # keep reference to the local identifier
        self.local_identifier = self.supvisors.mapper.local_identifier
        # initialize parameters
        self.parameters = {}
        self.store_message = None
        self.redirect = False
        # extract parameters from context
        self.update_strategy()
        self.update_auto_refresh()
        self.update_identifier()
        self.update_application_name()
        self.update_process_name()
        self.update_namespec()
        self.update_application_shrink_expand()
        self.update_process_shrink_expand()
        self.update_period()
        self.update_cpu_id()
        self.update_nic_name()
        self.update_disk_stats_choice()
        self.update_partition_name()
        self.update_device_name()

    # information extracted
    @property
    def strategy(self) -> StartingStrategies:
        return StartingStrategies[self.parameters[STRATEGY]]

    @property
    def auto_refresh(self) -> bool:
        return self.parameters[AUTO]

    @property
    def identifier(self) -> str:
        return self.parameters[IDENTIFIER]

    @property
    def application_name(self) -> str:
        return self.parameters[APPLI]

    @property
    def process_name(self) -> str:
        return self.parameters[PROCESS]

    @process_name.setter
    def process_name(self, proc_name: str) -> None:
        self.parameters[PROCESS] = proc_name

    @property
    def namespec(self) -> str:
        return self.parameters[NAMESPEC]

    @property
    def application_shex(self) -> str:
        return self.parameters[APP_SHRINK_EXPAND]

    @property
    def process_shex(self) -> str:
        return self.parameters[PROC_SHRINK_EXPAND]

    @property
    def period(self) -> float:
        return self.parameters[PERIOD]

    @property
    def cpu_id(self) -> int:
        return self.parameters[CPU]

    @property
    def nic_name(self) -> str:
        return self.parameters[NIC]

    @property
    def device(self) -> str:
        return self.parameters[DEVICE]

    @property
    def partition(self) -> str:
        return self.parameters[PARTITION]

    @property
    def disk_stats(self) -> str:
        return self.parameters[DISK_STATS]

    # simple extraction from context form
    @property
    def action(self) -> str:
        """ Extract action requested in context form. """
        return self.http_context.form.get(ACTION)

    @property
    def message(self) -> str:
        """ Extract message text from context form. """
        return self.http_context.form.get(MESSAGE)

    @property
    def gravity(self) -> str:
        """ Extract message gravity from context form. """
        return self.http_context.form.get(GRAVITY)

    def set_message_default(self, message: str, gravity: str):
        """ Set the message and gravity if there is no existing message. """
        if not self.message:
            self.http_context.form[GRAVITY] = gravity
            self.http_context.form[MESSAGE] = message

    # complex extraction from context form
    def update_period(self) -> None:
        """ Extract period from context. """
        default_value = next(iter(self.supvisors.options.stats_periods))
        self._update_float(PERIOD, self.supvisors.options.stats_periods, default_value)

    def update_strategy(self) -> None:
        """ Extract starting strategy from context. """
        self._update_string(STRATEGY, [x.name for x in StartingStrategies],
                            self.supvisors.options.starting_strategy.name)

    def update_auto_refresh(self) -> None:
        """ Extract auto refresh from context. """
        # assign value found or default
        self._update_boolean(AUTO, False)

    def update_identifier(self) -> None:
        """ Extract identifier from context. """
        # assign value found or default
        self._update_string(IDENTIFIER, self.supvisors.mapper.instances, self.local_identifier)

    def update_application_name(self) -> None:
        """ Extract application name from context. """
        self._update_string(APPLI, list(self.supvisors.context.applications.keys()))

    def update_process_name(self) -> None:
        """ Extract process name from context.
        ApplicationView may select a process unknown to this Supvisors instance. """
        status = self.supvisors.context.instances[self.parameters[IDENTIFIER]]
        # consider processes running on this Supvisors instance + supervisord
        running_processes = [x.namespec for x in status.running_processes()]
        running_processes.append('supervisord')
        self._update_string(PROCESS, running_processes)

    def update_namespec(self) -> None:
        """ Extract namespec from context. """
        value = None
        str_value = self.http_context.form.get(NAMESPEC)
        if str_value:
            # check that value is known to Supvisors
            if self.supvisors.context.is_namespec(str_value):
                value = str_value
            else:
                self.store_message = error_message(f'Incorrect {NAMESPEC}: {str_value}')
        # assign value found or default
        self.logger.trace(f'ViewContext.update_namespec: {NAMESPEC} set to {value}')
        self.parameters[NAMESPEC] = value

    def update_cpu_id(self) -> None:
        """ Extract CPU id from context. """
        self._update_integer(CPU, list(range(self.get_nb_cores(self.local_identifier) + 1)))

    def update_nic_name(self) -> None:
        """ Extract network interface name from context.
        Only the HostInstanceView displays interface data, so it's local. """
        stats_instance = self.get_instance_stats()
        interfaces = stats_instance.net_io.keys() if stats_instance else []
        default_value = next(iter(interfaces), None)
        self._update_string(NIC, interfaces, default_value)

    def update_disk_stats_choice(self) -> None:
        """ Extract the disk stats choice. """
        self._update_string(DISK_STATS, ['usage', 'io'], 'io')

    def update_partition_name(self) -> None:
        """ Extract interface name from context.
        Only the HostInstanceView displays interface data, so it's local. """
        stats_instance = self.get_instance_stats()
        partitions = stats_instance.disk_usage.keys() if stats_instance else []
        default_value = next(iter(partitions), None)
        self._update_string(PARTITION, partitions, default_value)

    def update_device_name(self) -> None:
        """ Extract interface name from context.
        Only the HostInstanceView displays interface data, so it's local. """
        stats_instance = self.get_instance_stats()
        devices = stats_instance.disk_io.keys() if stats_instance else []
        default_value = next(iter(devices), None)
        self._update_string(DEVICE, devices, default_value)

    @staticmethod
    def _get_default_shex(nb_items: int, expanded: bool) -> bytearray:
        """ Get a default shex bytearray filled with 1 if expanded.

        :param nb_items: the number of items in the shex.
        :param expanded: a status telling if the bytearray should be filled with 0 or 1.
        :return: the application shex bytearray.
        """
        base_value = 0xff if expanded else 0
        return bytearray([base_value] * math.ceil(nb_items / 8))

    def get_default_application_shex(self, expanded: bool) -> bytearray:
        """ Get a default application shex bytearray filled with 1 if expanded.

        :param expanded: a status telling if the bytearray should be filled with 0 or 1.
        :return: the application shex bytearray.
        """
        nb_applications = len(self.supvisors.context.applications)
        return ViewContext._get_default_shex(nb_applications, expanded)

    def get_default_process_shex(self, application_name: str, expanded: bool) -> bytearray:
        """ Get a default process shex bytearray filled with 1 if expanded, for the considered application.

        :param application_name: the application name to get the processes from.
        :param expanded: a status telling if the bytearray should be filled with 0 or 1.
        :return: the process shex bytearray.
        """
        nb_processes = len(self.supvisors.context.applications[application_name].processes)
        return ViewContext._get_default_shex(nb_processes, expanded)

    def update_application_shrink_expand(self):
        """ Extract process display choices from context.
        By default, all are displayed. """
        ba = self.get_default_application_shex(True)
        self._update_shex(APP_SHRINK_EXPAND, ba)

    def update_process_shrink_expand(self):
        """ Extract instance process display choices from context.
        By default, all are hidden. """
        if self.application_name:
            ba = self.get_default_process_shex(self.application_name, False)
            self._update_shex(PROC_SHRINK_EXPAND, ba)

    def _update_shex(self, param: str, default_ba: bytearray) -> None:
        """ Extract shex from context. """
        # extract mask from context
        str_value = self.http_context.form.get(param)
        if str_value:
            # check that value has correct format (only hex and size twice the size of the bytearray)
            try:
                value = bytearray.fromhex(str_value)
            except ValueError:
                self.logger.error(f'ViewContext._update_shex: non-hexadecimal {param}')
            else:
                if len(default_ba) != len(value):
                    self.logger.error(f'ViewContext._update_shex: {param} does not fit with the number of items')
                else:
                    default_ba = value
        self.logger.debug(f'ViewContext._update_shex: {param} set to {default_ba.hex()}')
        self.parameters[param] = default_ba.hex()

    def _update_string(self, param: str, check_list: NameList, default_value: str = None):
        """ Extract information from context based on allowed values in check_list. """
        value = default_value
        str_value = self.http_context.form.get(param)
        if str_value:
            # check that value is known to Supvisors
            if str_value in check_list:
                value = str_value
            else:
                self.store_message = error_message(f'Incorrect {param}: {str_value}')
        # assign value found or default
        self.logger.trace(f'ViewContext._update_string: {param} set to {value}')
        self.parameters[param] = value

    def _update_integer(self, param, check_list, default_value=0):
        """ Extract information from context and convert to integer based on allowed values in check_list. """
        value = default_value
        str_value = self.http_context.form.get(param)
        if str_value:
            try:
                int_value = int(str_value)
            except ValueError:
                self.store_message = error_message(f'{param} is not an integer: {str_value}')
            else:
                # check that int_value is defined in check list
                if int_value in check_list:
                    value = int_value
                else:
                    self.store_message = error_message(f'Incorrect {param}: {int_value}')
        # assign value found or default
        self.logger.trace(f'ViewContext._update_integer: {param} set to {value}')
        self.parameters[param] = value

    def _update_float(self, param, check_list, default_value=0.0):
        """ Extract information from context and convert to float based on allowed values in check_list. """
        value = default_value
        str_value = self.http_context.form.get(param)
        if str_value:
            try:
                float_value = float(str_value)
            except ValueError:
                self.store_message = error_message(f'{param} is not a float: {str_value}')
            else:
                # check that float_value is defined in check list
                if float_value in check_list:
                    value = float_value
                else:
                    self.store_message = error_message(f'Incorrect {param}: {float_value}')
        # assign value found or default
        self.logger.trace(f'ViewContext._update_float: {param} set to {value}')
        self.parameters[param] = value

    def _update_boolean(self, param, default_value=False):
        """ Extract information from context and convert to boolean based on allowed values in check_list. """
        value = default_value
        str_value = self.http_context.form.get(param)
        if str_value:
            try:
                value = strtobool(str_value)
            except ValueError:
                self.store_message = error_message(f'{param} is not a boolean-like: {str_value}')
        # assign value found or default
        self.logger.trace(f'ViewContext._update_boolean: {param} set to {value}')
        self.parameters[param] = value

    # URL formatting
    def url_parameters(self, reset_shex, **kwargs):
        """ Return the list of parameters for a URL. """
        parameters = dict(self.parameters, **kwargs)
        if reset_shex:
            del parameters[APP_SHRINK_EXPAND]
        return '&'.join([f'{key}={quote(str(value))}'
                         for key, value in sorted(parameters.items()) if value])

    def format_url(self, identifier: str, page: str, **kwargs):
        """ Format URL from parameters. """
        netloc = ''
        # build network location if identifier is provided
        if identifier:
            instance = self.supvisors.mapper.instances[identifier]
            netloc = f'http://{quote(instance.host_id)}:{instance.http_port}/'
        # shex must be reset if the Supvisors instance changes
        local_identifier = not identifier or identifier == self.local_identifier
        # build URL from netloc, page and attributes
        return f'{netloc}{page}?{self.url_parameters(not local_identifier, **kwargs)}'

    def fire_message(self) -> None:
        """ Set message in context response to be displayed at next refresh. """
        if self.store_message:
            args = {MESSAGE: self.store_message[1], GRAVITY: self.store_message[0]}
            form = self.http_context.form
            # if redirect requested, go back to main page
            path_translated = '/' + SUPVISORS_PAGE if self.redirect else form[PATH_TRANSLATED]
            location = f'{form[SERVER_URL]}{path_translated}?{self.url_parameters(False, **args)}'
            self.http_context.response[HEADERS][LOCATION] = location

    # Statistics
    def get_nb_cores(self, identifier: str) -> int:
        """ Get the number of processors of the host where the Supvisors instance is running. """
        # 2 chances to get the value
        nb_cores = self.supvisors.host_compiler.get_nb_cores(identifier)
        if not nb_cores:
            nb_cores = self.supvisors.process_compiler.get_nb_cores(identifier)
        return nb_cores

    def get_node_characteristics(self):
        """ Get the node characteristics from the stats collector. """
        if self.supvisors.stats_collector:
            return self.supvisors.stats_collector.node_info
        return None

    def get_instance_stats(self, identifier: str = None):
        """ Get the statistics structure related to the identifier and the period selected.
        If no identifier is specified, local identifier is used. """
        stats_identifier = identifier or self.local_identifier
        period = self.parameters.get(PERIOD)
        return self.supvisors.host_compiler.get_stats(stats_identifier, period)

    def get_process_stats(self, namespec: str, identifier: str):
        """ Get the statistics structure related to the process and the period selected.
        Get also the number of cores available where this Supvisors instance runs (useful for process CPU IRIX mode).
        """
        # get the number of cores for Solaris mode
        nb_cores = self.get_nb_cores(identifier)
        # return the process statistics for this process
        period = self.parameters.get(PERIOD)
        stats_instance = self.supvisors.process_compiler.get_stats(namespec, identifier, period)
        return nb_cores, stats_instance

    def get_process_status(self, namespec: str = None) -> Optional[ProcessStatus]:
        """ Get the ProcessStatus instance related to the process named namespec.
        If none specified, the form namespec is used. """
        namespec = namespec or self.namespec
        if namespec:
            try:
                return self.supvisors.context.get_process(namespec)
            except KeyError:
                self.logger.debug(f'ViewContext.get_process_status: failed to get ProcessStatus from {namespec}')

    # shex access
    def get_application_shex(self, application_name: str) -> Tuple[bool, str]:
        """ Get the expand / shrink value of the application and the shex string to invert it.

        :param application_name: the application name.
        :return: the application shex and the inverted shex.
        """
        ba = bytearray.fromhex(self.application_shex)
        # get the index of the application in context
        idx = sorted(self.supvisors.context.applications).index(application_name)
        # get application shex value
        application_shex = bool(get_bit(ba, idx))
        # get new shex with inverted value for application
        set_bit(ba, idx, not application_shex)
        return application_shex, ba.hex()

    def get_process_shex(self, process_name: str) -> Tuple[bool, str]:
        """ Get the expand / shrink value of the process and the shex string to invert it.

        :param process_name: the process name.
        :return: the process shex and the inverted shex.
        """
        ba = bytearray.fromhex(self.process_shex)
        # get the index of the application in context
        idx = sorted(self.supvisors.context.applications[self.application_name].processes).index(process_name)
        # get application shex value
        process_shex = bool(get_bit(ba, idx))
        # get new shex with inverted value for application
        set_bit(ba, idx, not process_shex)
        return process_shex, ba.hex()

    @staticmethod
    def cpu_id_to_string(idx):
        """ Get a printable form of cpu index. """
        return f'{idx - 1}' if idx > 0 else 'all'



# ======================================================================
# Copyright 2016 Julien LE CLEACH
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import re
from collections import OrderedDict
from socket import getfqdn, gethostbyaddr, gethostname, herror, gaierror
from typing import Any, Dict, Optional, Tuple

from supervisor.loggers import Logger

from supvisors.ttypes import Ipv4Address, NameList, NameSet, Payload

# default identifier given by Supervisor
DEFAULT_IDENTIFIER = 'supervisor'


def get_addresses(host_id: str, logger: Logger) -> Optional[Tuple[str, NameList, NameList]]:
    """ Get hostname, aliases and all IP addresses for the host_id.

    :param host_id: the host_name used in the Supvisors option
    :param logger: the Supvisors logger
    :return: the list of possible node name or IP addresses
    """
    try:
        return gethostbyaddr(host_id)  # hostname, aliases, ip_addresses
    except (herror, gaierror):
        logger.error(f'get_addresses: unknown address {host_id}')


class SupvisorsInstanceId:
    """ Identification attributes for a Supvisors instance.

    The Supvisors instance uses the Supervisor identifier if provided, or builds an identifier from the host name and
    the HTTP port used by this Supervisor.

    Attributes are:
        - identifier: the Supvisors identifier 'host_name:http_port' ;
        - nick_identifier: the Supervisor identifier (defaulted to identifier) if not provided ;
        - host_id: the host where the Supervisor is running ;
        - http_port: the Supervisor port number ;
        - event_port: the port number used to publish all Supvisors events ;
        - host_name: the host name corresponding to the host id, as known by the local network configuration ;
        - ip_address: the main ip_address of the host, as known by the local network configuration ;
        - stereotype: the stereotype (used for rules).
    """

    PATTERN = re.compile(r'^(<(?P<identifier>[\w\-:.]+)>)?(?P<host>[\w\-.]+)(:(?P<http_port>\d{4,5}))?$')

    # attributes taken from input
    identifier = None
    nick_identifier = None
    host_id = None
    http_port = None
    event_port = None
    # attributes got from network
    host_name = None
    aliases = None
    ip_addresses = None

    def __init__(self, item: str, supvisors: Any):
        """ Initialization of the attributes.

        :param item: the Supervisor parameters to be parsed
        """
        self.supvisors = supvisors
        # attributes set a posteriori
        self.stereotypes: NameList = []
        # parse item to get the values
        self.parse_from_string(item)
        self.check_values()
        # choose the node name among the possible node identifiers
        if self.host_id:
            addresses = get_addresses(self.host_id, self.logger)
            if addresses:
                self.host_name, self.aliases, self.ip_addresses = addresses
                self.logger.debug(f'SupvisorsInstanceId: host_id={self.host_id} host_name={self.host_name}'
                                  f' aliases={self.aliases} ip_addresses={self.ip_addresses}')

    @property
    def logger(self) -> Logger:
        """ Shortcut to the Supvisors logger. """
        return self.supvisors.logger

    @property
    def ip_address(self) -> Optional[str]:
        """ Return the main IP address. """
        return self.ip_addresses[0] if self.ip_addresses else None

    @property
    def source(self) -> Tuple[str, str, Ipv4Address]:
        """ Return the identification details of the instance. """
        return self.identifier, self.nick_identifier, (self.ip_address, self.http_port)

    def host_matches(self, fdqn: str) -> bool:
        """ Check if the fully-qualified domain name matches the host name or any alias.

        :param fdqn: the fully-qualified domain name to check
        :return: True if fdqn is the local host name or a known alias
        """
        return fdqn == self.host_name or fdqn in self.aliases

    def check_values(self) -> None:
        """ Complete information where not provided.

        :return: None
        """
        http_port = self.http_port or self.supvisors.supervisor_data.server_port
        # define common identifier
        if self.host_id:
            self.identifier = f'{self.host_id}:{http_port}'
            # if nick_identifier is not set or default Supervisor identifier is used, assign an automatic identifier
            if not self.nick_identifier or self.nick_identifier == DEFAULT_IDENTIFIER:
                if self.http_port:
                    self.nick_identifier = self.identifier
                else:
                    self.nick_identifier = self.host_id
        # if http_port is not provided, use the local http_port value
        if not self.http_port:
            self.http_port = http_port
        # define event_port using option value if set
        if self.supvisors.options.event_port:
            self.event_port = self.supvisors.options.event_port
        else:
            # by default, assign to http_port + 1
            self.event_port = self.http_port + 1
        self.logger.debug(f'SupvisorsInstanceId.check_values: identifier={self.identifier}'
                          f' nick_identifier={self.nick_identifier}'
                          f' host_id={self.host_id} http_port={self.http_port}')

    def parse_from_string(self, item: str):
        """ Parse string according to PATTERN to get the Supvisors instance identification attributes.

        :param item: the parameters to be parsed
        :return: None
        """
        pattern_match = SupvisorsInstanceId.PATTERN.match(item)
        if pattern_match:
            self.nick_identifier = pattern_match.group('identifier')
            self.host_id = pattern_match.group('host')
            # check http port
            port = pattern_match.group('http_port')
            self.http_port = int(port) if port else 0
            self.logger.debug(f'SupvisorsInstanceId.parse_from_string: identifier={self.identifier}'
                              f' host_id={self.host_id} http_port={self.http_port}')

    def serial(self) -> Payload:
        """ Get a serializable form of the instance parameters.

        :return: the instance parameters in a dictionary.
        """
        return {'identifier': self.identifier,
                'nick_identifier': self.nick_identifier,
                'host_id': self.host_id,
                'http_port': self.http_port,
                'host_name': self.host_name,
                'ip_addresses': self.ip_addresses,
                'stereotypes': self.stereotypes}

    def __repr__(self) -> str:
        """ Initialization of the attributes.

        :return: the identifier as string representation of the SupvisorsInstanceId object.
        """
        if self.identifier == self.nick_identifier:
            return self.identifier
        return f'<{self.nick_identifier}>{self.identifier}'


class SupvisorsMapper:
    """ Class used for storage of the Supvisors instances declared in the configuration file.

    Attributes are:
        - supvisors: the global Supvisors structure ;
        - logger: the reference to the common logger ;
        - _instances: the list of Supvisors instances declared in the supvisors section of the Supervisor
          configuration file ;
        - _nick_identifiers: the list of nick identifiers that can be used on all Supvisors user interfaces ;
        - _nodes: the list of Supvisors instances grouped by node names ;
        - _core_identifiers: the list of Supvisors core identifiers declared in the supvisors section of the Supervisor
          configuration file ;
        - local_identifier: the local Supvisors identifier ;
        - initial_identifiers: the initial list of Supvisors instances (excluding the discovered ones) ;
        - stereotypes: the Supvisors identifiers, sorted by stereotype.
    """

    # annotation types
    InstancesMap = Dict[str, SupvisorsInstanceId]

    def __init__(self, supvisors: Any):
        """ Initialization of the attributes.

        :param supvisors: the global Supvisors structure.
        """
        self.supvisors = supvisors
        self._instances: SupvisorsMapper.InstancesMap = OrderedDict()
        self._nick_identifiers: Dict[str, str] = {}  # {nick_identifier: identifier}
        self._nodes: Dict[str, NameList] = {}
        self._core_identifiers: NameList = []
        self.local_identifier: Optional[str] = None
        self.initial_identifiers: NameList = []
        self.stereotypes: Dict[str, NameList] = {}

    @property
    def logger(self) -> Logger:
        """ Return the Supvisors logger. """
        return self.supvisors.logger

    @property
    def local_instance(self) -> SupvisorsInstanceId:
        """ Property getter for the SupvisorsInstanceId corresponding to local_identifier.

        :return: the local SupvisorsInstanceId.
        """
        return self._instances[self.local_identifier]

    @property
    def local_nick_identifier(self) -> str:
        """ Property getter for the nick identifier of the local Supvisors instance.

        :return: the local nick identifier.
        """
        return self.local_instance.nick_identifier

    @property
    def instances(self) -> InstancesMap:
        """ Property getter for the _instances attribute.

        :return: the list of Supvisors instances configured in Supvisors
        """
        return self._instances

    @property
    def nodes(self) -> Dict[str, NameList]:
        """ Property getter for the _nodes attribute.

        :return: the Supvisors identifiers per node
        """
        return self._nodes

    @property
    def core_identifiers(self) -> NameList:
        """ Get the known Supvisors core identifiers.

        :return: the minimum Supvisors identifiers to end the synchronization phase
        """
        return self.filter(self._core_identifiers)

    def get_nick_identifier(self, identifier: str) -> str:
        """ Property getter for the nick identifier of the Supvisors instance.

        :return: the nick identifier.
        """
        return self.instances[identifier].nick_identifier

    def add_instance(self, item: str) -> SupvisorsInstanceId:
        """ Store a new Supvisors instance using a format compliant with the supvisors_list option.

        :param item: the Supvisors instance to add.
        :return: the new Supvisors instance.
        """
        supvisors_id = SupvisorsInstanceId(item, self.supvisors)
        if supvisors_id.ip_address:
            self._instances[supvisors_id.identifier] = supvisors_id
            self._nick_identifiers[supvisors_id.nick_identifier] = supvisors_id.identifier
            self._nodes.setdefault(supvisors_id.ip_address, []).append(supvisors_id.identifier)
            return supvisors_id
        message = f'could not define a Supvisors identification from "{item}"'
        raise ValueError(message)

    def configure(self, supvisors_list: NameList, stereotypes: NameSet, core_list: NameList) -> None:
        """ Store the identification of the Supvisors instances declared in the configuration file and determine
        the local Supvisors instance in this list.

        :param supvisors_list: the Supvisors instances declared in the supvisors section of the configuration file.
        :param stereotypes: the local Supvisors instance stereotypes.
        :param core_list: the minimum Supvisors identifiers to end the synchronization phase.
        :return: None
        """
        if supvisors_list:
            # get Supervisor identification from each element
            for item in supvisors_list:
                self.add_instance(item)
            # keep information about the initial Supvisors identifiers added to the configuration
            self.initial_identifiers = list(self._instances.keys())
        else:
            # if supvisors_list is empty, use self identification from supervisor internal data
            supervisor = self.supvisors.supervisor_data
            item = f'<{supervisor.identifier}>{gethostname()}:{supervisor.server_port}'
            self.logger.info(f'SupvisorsMapper.configure: define local Supvisors as {item}')
            self.add_instance(item)
        self.logger.info(f'SupvisorsMapper.configure: identifiers={self._nick_identifiers}')
        self.logger.info(f'SupvisorsMapper.configure: nodes={self.nodes}')
        # get local Supervisor identification from list
        self.find_local_identifier(stereotypes)
        # NOTE: store the core identifiers without filtering as is.
        #       they can only be filtered on access because of discovery mode.
        self._core_identifiers = core_list
        self.logger.info(f'SupvisorsMapper.configure: core_identifiers={core_list}')

    def find_local_identifier(self, stereotypes: NameSet):
        """ Find the local Supvisors identification in the list declared in the configuration file.
        It can be either the local Supervisor identifier or a name built using the local host name or any of its known
        IP addresses.

        :param stereotypes: the local Supvisors instance stereotypes.
        :return: the identifier of the local Supvisors.
        """
        # get the must-match parameters
        local_host_name = getfqdn()
        local_http_port = self.supvisors.supervisor_data.server_port
        # try to find a Supvisors instance corresponding to the local configuration
        # WARN: there MUST be exactly one unique matching Supvisors instance
        matching_identifiers = [sup_id.identifier
                                for sup_id in self._instances.values()
                                if sup_id.host_matches(local_host_name) and sup_id.http_port == local_http_port]
        if len(matching_identifiers) == 1:
            self.local_identifier = matching_identifiers[0]
            if ((self.supvisors.supervisor_data.identifier != DEFAULT_IDENTIFIER)
                    and (self.local_nick_identifier != self.supvisors.supervisor_data.identifier)):
                self.logger.warn('SupvisorsMapper.find_local_identifier: mismatch between Supervisor identifier'
                                 f' "{self.supvisors.supervisor_data.identifier}"'
                                 f' and local Supvisors in supvisors_list "{self.local_nick_identifier}"')
            # assign the generic Supvisors instance stereotype
            self.assign_stereotypes(self.local_identifier, stereotypes)
        else:
            if len(matching_identifiers) > 1:
                message = f'multiple candidates for the local Supvisors identifiers={matching_identifiers}'
            else:
                message = 'could not find the local Supvisors in supvisors_list'
            self.logger.error(f'SupvisorsMapper.find_local_identifier: {message}')
            raise ValueError(message)
        self.logger.info(f'SupvisorsMapper.find_local_identifier: {str(self.local_instance)}')

    def assign_stereotypes(self, identifier: str, stereotypes: NameSet) -> None:
        """ Assign stereotypes to the Supvisors instance.

        The method maintains a map of Supvisors instances per stereotype.
        The list of Supvisors instances per stereotype is ordered the same way as the Supvisors instances themselves.

        :param identifier: the identifier of the Supvisors instance.
        :param stereotypes: the stereotypes of the Supvisors instance.
        :return: None
        """
        # assign stereotypes on the first attempt only
        sup_id: SupvisorsInstanceId = self.instances[identifier]
        if stereotypes and not sup_id.stereotypes:
            self.logger.info(f'SupvisorsMapper.assign_stereotype: identifier={identifier} stereotypes={stereotypes}')
            # set stereotype in SupvisorsInstanceId
            sup_id.stereotypes = list(stereotypes)
            # update map
            for stereotype in stereotypes:
                if stereotype in self.stereotypes:
                    # a new identifier has to be added to the list, keeping the same order as in self._instances
                    identifiers = self.stereotypes[stereotype]
                    identifiers.append(identifier)
                    self.stereotypes[stereotype] = [x for x in self.instances.keys() if x in identifiers]
                else:
                    self.stereotypes[stereotype] = [identifier]

    def filter(self, identifier_list: NameList) -> NameList:
        """ Check the identifiers list against the identifiers declared in the configuration file
        or via the discovery mode.

        The identifier list can be made of Supvisors identifiers, Supervisor identifiers and/or stereotypes.

        If the identifier is not found, it is removed from the list.
        If more than one occurrence of the same identifier is found, only the first one is kept.

        :param identifier_list: a list of Supvisors identifiers and/or nick identifiers and/or stereotypes.
        :return: the filtered list of Supvisors identifiers.
        """
        identifiers = []
        # filter unknown Supvisors identifiers and expand the stereotypes
        for identifier in identifier_list:
            if identifier in self._instances:
                identifiers.append(identifier)
            elif identifier in self._nick_identifiers:
                identifiers.append(self._nick_identifiers[identifier])
            elif identifier in self.stereotypes:
                # identifier is a stereotype
                identifiers.extend(self.stereotypes[identifier])
            else:
                self.logger.warn(f'SupvisorsMapper.filter: identifier={identifier} invalid')
        # remove duplicates keeping the same order
        return list(OrderedDict.fromkeys(identifiers))

    def check_candidate(self, identifier: str, nick_identifier: str) -> bool:
        """ Return True if the candidate is eligible as a new Supvisors instance.

        :param identifier: the candidate identifier.
        :param nick_identifier: the candidate nick identifier.
        :return: True if the candidate is unknown.
        """
        if nick_identifier in self._nick_identifiers:
            known_identifier = self._nick_identifiers[nick_identifier]
            if known_identifier != identifier:
                self.logger.warn(f'SupvisorsMapper.check_candidate: candidate=<{nick_identifier}>{identifier}'
                                 f' incompatible with the Supvisors instance known as'
                                 f' <{nick_identifier}>{known_identifier}')
            return False
        if identifier in self.instances:
            known_nick_identifier = self.get_nick_identifier(identifier)
            if known_nick_identifier != nick_identifier:
                self.logger.warn(f'SupvisorsMapper.check_candidate: candidate=<{nick_identifier}>{identifier}'
                                 f' incompatible with the Supvisors instance known as'
                                 f' <{known_nick_identifier}>{identifier}')
            return False
        return True



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import io
import os

import supervisor
from supervisor.compat import as_bytes
from supervisor.medusa.http_server import http_date


# exchange class for images
class StatsImage:
    """ Buffer class holding PNG contents. """

    def __init__(self):
        self.contents = None

    def new_image(self):
        if self.contents:
            self.contents.close()
        self.contents = io.BytesIO()
        return self.contents


# instance for image buffers
host_cpu_img = StatsImage()
host_mem_img = StatsImage()
host_net_io_img = StatsImage()
host_disk_io_img = StatsImage()
host_disk_usage_img = StatsImage()

process_cpu_img = StatsImage()
process_mem_img = StatsImage()


# simple handlers for web images
class ImageView:
    content_type = 'image/png'
    delay = .5

    def __init__(self, context, buffer):
        self.context = context
        self.buffer = buffer

    def __call__(self):
        response = self.context.response
        headers = response['headers']
        headers['Content-Type'] = self.content_type
        headers['Pragma'] = 'no-cache'
        headers['Cache-Control'] = 'no-cache'
        headers['Expires'] = http_date.build_http_date(0)
        if self.buffer.contents:
            body = self.buffer.contents.getvalue()
        else:
            body = b""
        response['body'] = as_bytes(body)
        return response


class HostCpuImageView(ImageView):
    """ Dummy view holding the Host CPU image. """

    def __init__(self, context):
        """ Link to the Host CPU buffer. """
        ImageView.__init__(self, context, host_cpu_img)


class HostMemoryImageView(ImageView):
    """ Dummy view holding the Host Memory image. """

    def __init__(self, context):
        """ Link to the Host Memory buffer. """
        ImageView.__init__(self, context, host_mem_img)


class HostNetworkIoImageView(ImageView):
    """ Dummy view holding the Host Network IO image. """

    def __init__(self, context):
        """ Link to the Host Network IO buffer. """
        ImageView.__init__(self, context, host_net_io_img)


class HostDiskIoImageView(ImageView):
    """ Dummy view holding the Host Disk IO image. """

    def __init__(self, context):
        """ Link to the Host Network buffer. """
        ImageView.__init__(self, context, host_disk_io_img)


class HostDiskUsageImageView(ImageView):
    """ Dummy view holding the Host Disk usage image. """

    def __init__(self, context):
        """ Link to the Host Disk usage buffer. """
        ImageView.__init__(self, context, host_disk_usage_img)


class ProcessCpuImageView(ImageView):
    """ Dummy view holding the Process CPU image. """

    def __init__(self, context):
        """ Link to the Process CPU buffer. """
        ImageView.__init__(self, context, process_cpu_img)


class ProcessMemoryImageView(ImageView):
    """ Dummy view holding the Process Memory image. """

    def __init__(self, context):
        """ Link to the Process Memory buffer. """
        ImageView.__init__(self, context, process_mem_img)


# Trick to make available the Supervisor icon from Supvisors Web UI
# There might be a better way in medusa but TBD
def create_icon(icon_path) -> StatsImage:
    """ Load an image into a BytesIO. """
    icon = StatsImage()
    with io.open(icon_path, 'rb') as fic:
        icon.contents = io.BytesIO(fic.read())
    return icon


class SupervisorIconImage(ImageView):
    """ Dummy view holding the Supervisor Icon. """

    _icon: StatsImage = None

    def __init__(self, context):
        """ Link to the Process Memory buffer. """
        ImageView.__init__(self, context, self.icon)

    @property
    def icon(self) -> StatsImage:
        """ Load the Supervisor icon into a BytesIO. """
        if not SupervisorIconImage._icon:
            icon_path = os.path.join(os.path.dirname(supervisor.__file__), 'ui', 'images', 'icon.png')
            SupervisorIconImage._icon = create_icon(icon_path)
        return SupervisorIconImage._icon


class SoftwareIconImage(ImageView):
    """ Dummy view holding the Process Memory image. """

    _icon: StatsImage = None

    def __init__(self, context):
        """ Link to the Process Memory buffer. """
        ImageView.__init__(self, context, self._icon)

    @staticmethod
    def set_path(icon_path: str) -> None:
        """ Load the user software icon into a BytesIO. """
        if not SoftwareIconImage._icon and icon_path:
            SoftwareIconImage._icon = create_icon(icon_path)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# ======================================================================
# Copyright 2022 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from flask_restx import Api
from supervisor.compat import xmlrpclib

from .system_namespace import api as system_api
from .supervisor_namespace import api as supervisor_api
from .supvisors_namespace import api as supvisors_api

# create Api with all namespaces
api = Api(title='Supvisors Flask interface')
api.add_namespace(system_api)
api.add_namespace(supervisor_api)
api.add_namespace(supvisors_api)


@api.errorhandler
def default_error_handler(error):
    """ Default error handler. """
    return {'message': str(error)}, getattr(error, 'code', 500)


@api.errorhandler(xmlrpclib.Fault)
def supervisor_error_handler(error):
    """ Supervisor error handler. """
    return {'message': error.faultString, 'code': error.faultCode}, 400




#!/usr/bin/python
# -*- coding: utf-8 -*-

# ======================================================================
# Copyright 2022 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from flask import g, jsonify
from flask_restx import Namespace, Resource, fields, inputs

from supvisors.rpcinterface import RPCInterface
from supvisors.ttypes import ConciliationStrategies, StartingStrategies
from .utils import get_docstring_description, get_docstring_parameters

# Utilities
StartingStrategiesParam = ', '.join([x.name for x in StartingStrategies])
ConciliationStrategiesParam = ', '.join([x.name for x in ConciliationStrategies])
LoggerLevelsParam = ', '.join(RPCInterface.get_logger_levels().values())

# Supvisors part
api = Namespace('supvisors', description='Supvisors operations')


# Request parsers
wait_parser = api.parser()
wait_parser.add_argument('wait', type=inputs.boolean, default=True,
                         help='if ``True``, wait until completion of the request')

start_process_parser = api.parser()
start_process_parser.add_argument('extra_args', type=str, default='',
                                  help='the extra arguments to be passed to the command line of the program')
start_process_parser.add_argument('wait', type=inputs.boolean, default=True,
                                  help='if ``True``, wait until completion of the request')

# Models
wait_model = api.model('Wait', {'wait': fields.Boolean})
start_process_model = api.clone('Start process', wait_model, {'extra_args': fields.String})


# Routes
@api.route('/api_version', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_api_version))
class SupvisorsApiVersion(Resource):
    @staticmethod
    def get():
        return g.proxy.supvisors.get_api_version()


@api.route('/supvisors_state', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_supvisors_state))
class SupvisorsState(Resource):
    @staticmethod
    def get():
        return g.proxy.supvisors.get_supvisors_state()


@api.route('/master_identifier', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_master_identifier))
class SupvisorsMasterIdentifier(Resource):
    @staticmethod
    def get():
        return g.proxy.supvisors.get_master_identifier()


@api.route('/strategies', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_strategies))
class SupvisorsStrategies(Resource):
    @staticmethod
    def get():
        return g.proxy.supvisors.get_strategies()


@api.route('/all_instances_info', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_all_instances_info))
class SupvisorsAllInstancesInfo(Resource):
    @staticmethod
    def get():
        return jsonify(g.proxy.supvisors.get_all_instances_info())


@api.route('/instance_info/<string:identifier>', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_instance_info))
class SupvisorsInstanceInfo(Resource):
    @api.doc(params=get_docstring_parameters(RPCInterface.get_instance_info))
    def get(self, identifier):
        return g.proxy.supvisors.get_instance_info(identifier)


@api.route('/all_applications_info', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_all_applications_info))
class SupvisorsAllApplicationsInfo(Resource):
    @staticmethod
    def get():
        return jsonify(g.proxy.supvisors.get_all_applications_info())


@api.route('/application_info/<string:application_name>', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_application_info))
class SupvisorsApplicationInfo(Resource):
    @api.doc(params=get_docstring_parameters(RPCInterface.get_application_info))
    def get(self, application_name):
        return g.proxy.supvisors.get_application_info(application_name)


@api.route('/application_rules/<string:application_name>', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_application_rules))
class SupvisorsApplicationRules(Resource):
    @api.doc(params=get_docstring_parameters(RPCInterface.get_application_rules))
    def get(self, application_name):
        return g.proxy.supvisors.get_application_rules(application_name)


@api.route('/all_process_info', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_all_process_info))
class SupvisorsAllProcessInfo(Resource):
    @staticmethod
    def get():
        return jsonify(g.proxy.supvisors.get_all_process_info())


@api.route('/process_info/<string:namespec>', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_process_info))
class SupvisorsProcessInfo(Resource):
    @api.doc(params=get_docstring_parameters(RPCInterface.get_process_info))
    def get(self, namespec):
        return jsonify(g.proxy.supvisors.get_process_info(namespec))


@api.route('/all_local_process_info', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_all_local_process_info))
class SupvisorsAllLocalProcessInfo(Resource):
    @staticmethod
    def get():
        return jsonify(g.proxy.supvisors.get_all_local_process_info())


@api.route('/local_process_info/<string:namespec>', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_local_process_info))
class SupvisorsLocalProcessInfo(Resource):
    @api.doc(params=get_docstring_parameters(RPCInterface.get_local_process_info))
    def get(self, namespec):
        return g.proxy.supvisors.get_local_process_info(namespec)


@api.route('/process_rules/<string:namespec>', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_process_rules))
class SupvisorsProcessRules(Resource):
    @api.doc(params=get_docstring_parameters(RPCInterface.get_process_rules))
    def get(self, namespec):
        return jsonify(g.proxy.supvisors.get_process_rules(namespec))


@api.route('/conflicts', methods=('GET',))
@api.doc(description=get_docstring_description(RPCInterface.get_conflicts))
class SupvisorsConflicts(Resource):
    @staticmethod
    def get():
        return jsonify(g.proxy.supvisors.get_conflicts())


@api.route(f'/start_application/<any({StartingStrategiesParam}):strategy>/<string:application_name>', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.start_application))
class SupvisorsStartApplication(Resource):
    # keep this instead of using get_docstring_parameters as it is useful to display the enumeration literals
    @api.doc(params={'strategy': f'the starting strategy in {{{StartingStrategiesParam}}}',
                     'application_name': 'the name of the application'})
    @api.expect(wait_parser)
    def post(self, strategy, application_name):
        args = wait_parser.parse_args()
        return g.proxy.supvisors.start_application(strategy, application_name, args.wait)


@api.route(f'/stop_application/<string:application_name>', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.stop_application))
class SupvisorsStopApplication(Resource):
    @api.doc(params=get_docstring_parameters(RPCInterface.stop_application))
    @api.expect(wait_parser)
    def post(self, application_name):
        args = wait_parser.parse_args()
        return g.proxy.supvisors.stop_application(application_name, args.wait)


@api.route(f'/restart_application/<any({StartingStrategiesParam}):strategy>/<string:application_name>',
           methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.restart_application))
class SupvisorsRestartApplication(Resource):
    # keep this instead of using get_docstring_parameters as it is useful to display the enumeration literals
    @api.doc(params={'strategy': f'the starting strategy in {{{StartingStrategiesParam}}}',
                     'application_name': 'the name of the application'})
    @api.expect(wait_parser)
    def post(self, strategy, application_name):
        args = wait_parser.parse_args()
        return g.proxy.supvisors.restart_application(strategy, application_name, args.wait)


@api.route(f'/start_args/<string:namespec>', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.start_args))
class SupvisorsStartArgs(Resource):
    @api.doc(params=get_docstring_parameters(RPCInterface.start_args))
    @api.expect(start_process_parser)
    def post(self, namespec):
        args = start_process_parser.parse_args()
        return g.proxy.supvisors.start_args(namespec, args.extra_args, args.wait)


@api.route(f'/start_process/<any({StartingStrategiesParam}):strategy>/<string:namespec>', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.start_process))
class SupvisorsStartProcess(Resource):
    # keep this instead of using get_docstring_parameters as it is useful to display the enumeration literals
    @api.doc(params={'strategy': f'the starting strategy in {{{StartingStrategiesParam}}}',
                     'namespec': 'the namespec of the process to start'})
    @api.expect(start_process_parser)
    def post(self, strategy, namespec):
        args = start_process_parser.parse_args()
        return g.proxy.supvisors.start_process(strategy, namespec, args.extra_args, args.wait)


@api.route(f'/start_any_process/<any({StartingStrategiesParam}):strategy>/<string:regex>', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.start_any_process))
class SupvisorsStartAnyProcess(Resource):
    # keep this instead of using get_docstring_parameters as it is useful to display the enumeration literals
    @api.doc(params={'strategy': f'the starting strategy in {{{StartingStrategiesParam}}}',
                     'regex': 'the regular expression used to find a process to start'})
    @api.expect(start_process_parser)
    def post(self, strategy, regex):
        args = start_process_parser.parse_args()
        return g.proxy.supvisors.start_any_process(strategy, regex, args.extra_args, args.wait)


@api.route(f'/stop_process/<string:namespec>', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.stop_process))
class SupvisorsStopProcess(Resource):
    @api.doc(params=get_docstring_parameters(RPCInterface.stop_process))
    @api.expect(wait_parser)
    def post(self, namespec):
        args = wait_parser.parse_args()
        return g.proxy.supvisors.stop_process(namespec, args.wait)


@api.route(f'/restart_process/<any({StartingStrategiesParam}):strategy>/<string:namespec>', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.restart_process))
class SupvisorsRestartProcess(Resource):
    # keep this instead of using get_docstring_parameters as it is useful to display the enumeration literals
    @api.doc(params={'strategy': f'the starting strategy in {{{StartingStrategiesParam}}}',
                     'namespec': 'the namespec of the process to restart'})
    @api.expect(start_process_parser)
    def post(self, strategy, namespec):
        args = start_process_parser.parse_args()
        return g.proxy.supvisors.restart_process(strategy, namespec, args.extra_args, args.wait)


@api.route(f'/update_numprocs/<string:program_name>/<int:numprocs>', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.update_numprocs))
class SupvisorsUpdateNumprocs(Resource):
    @api.doc(params=get_docstring_parameters(RPCInterface.update_numprocs))
    @api.expect(wait_parser)
    def post(self, program_name, numprocs):
        args = wait_parser.parse_args()
        return g.proxy.supvisors.update_numprocs(program_name, numprocs, args.wait)


@api.route(f'/enable/<string:program_name>', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.enable))
class SupvisorsEnable(Resource):
    @api.doc(params=get_docstring_parameters(RPCInterface.enable))
    @api.expect(wait_parser)
    def post(self, program_name):
        args = wait_parser.parse_args()
        return g.proxy.supvisors.enable(program_name, args.wait)


@api.route(f'/disable/<string:program_name>', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.disable))
class SupvisorsDisable(Resource):
    @api.doc(params=get_docstring_parameters(RPCInterface.disable))
    @api.expect(wait_parser)
    def post(self, program_name):
        args = wait_parser.parse_args()
        return g.proxy.supvisors.disable(program_name, args.wait)


@api.route(f'/conciliate/<any({ConciliationStrategiesParam}):strategy>', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.conciliate))
class SupvisorsConciliate(Resource):
    # keep this instead of using get_docstring_parameters as it is useful to display the enumeration literals
    @api.doc(params={'strategy': f'the conciliation strategy in {{{ConciliationStrategiesParam}}}'})
    def post(self, strategy):
        return g.proxy.supvisors.conciliate(strategy)


@api.route(f'/restart_sequence', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.restart_sequence))
class SupvisorsRestartSequence(Resource):
    @api.expect(wait_parser)
    def post(self):
        args = wait_parser.parse_args()
        return g.proxy.supvisors.restart_sequence(args.wait)


@api.route(f'/restart', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.restart))
class SupvisorsRestart(Resource):
    @staticmethod
    def post():
        return g.proxy.supvisors.restart()


@api.route(f'/shutdown', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.shutdown))
class SupvisorsShutdown(Resource):
    @staticmethod
    def post():
        return g.proxy.supvisors.shutdown()


@api.route(f'/end_sync', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.end_sync))
class SupvisorsEndSync(Resource):
    @staticmethod
    def post():
        return g.proxy.supvisors.end_sync()


@api.route(f'/end_sync/<string:identifier>', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.end_sync))
class SupvisorsEndSyncMaster(Resource):
    @staticmethod
    def post(identifier: str):
        return g.proxy.supvisors.end_sync(identifier)


@api.route(f'/change_log_level/<any({LoggerLevelsParam}):log_level>', methods=('POST',))
@api.doc(description=get_docstring_description(RPCInterface.change_log_level))
class SupvisorsChangeLogLevel(Resource):
    # keep this instead of using get_docstring_parameters as it is useful to display the enumeration literals
    @api.doc(params={'log_level': f'the new logger level in {{{LoggerLevelsParam}}}'})
    def post(self, log_level):
        return g.proxy.supvisors.change_log_level(log_level)



# ======================================================================
# Copyright 2016 Julien LE CLEACH
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

from unittest.mock import Mock

import pytest

from supvisors.ttypes import *


def test_supvisors_instance_states():
    """ Test the SupvisorsInstanceStates enumeration. """
    expected = ['UNKNOWN', 'CHECKING', 'CHECKED', 'RUNNING', 'SILENT', 'ISOLATED']
    assert [x.name for x in SupvisorsInstanceStates] == expected


def test_supvisors_states():
    """ Test the SupvisorsStates enumeration. """
    expected = ['OFF', 'INITIALIZATION', 'DISTRIBUTION', 'OPERATION', 'CONCILIATION', 'RESTARTING',
                'SHUTTING_DOWN', 'FINAL']
    assert [x.name for x in SupvisorsStates] == expected


def test_application_states():
    """ Test the ApplicationStates enumeration. """
    expected = ['STOPPED', 'STARTING', 'RUNNING', 'STOPPING', 'DELETED']
    assert [x.name for x in ApplicationStates] == expected


def test_event_links():
    """ Test the EventLinks enumeration. """
    expected = ['NONE', 'ZMQ', 'WS']
    assert [x.name for x in EventLinks] == expected


def test_starting_strategies():
    """ Test the StartingStrategies enumeration. """
    expected = ['CONFIG', 'LESS_LOADED', 'MOST_LOADED', 'LOCAL', 'LESS_LOADED_NODE', 'MOST_LOADED_NODE']
    assert [x.name for x in StartingStrategies] == expected


def test_conciliation_strategies():
    """ Test the ConciliationStrategies enumeration. """
    expected = ['SENICIDE', 'INFANTICIDE', 'USER', 'STOP', 'RESTART', 'RUNNING_FAILURE']
    assert [x.name for x in ConciliationStrategies] == expected


def test_starting_failure_strategies():
    """ Test the StartingFailureStrategies enumeration. """
    expected = ['ABORT', 'STOP', 'CONTINUE']
    assert [x.name for x in StartingFailureStrategies] == expected


def test_running_failure_strategies():
    """ Test the RunningFailureStrategies enumeration. """
    expected = ['CONTINUE', 'RESTART_PROCESS', 'STOP_APPLICATION', 'RESTART_APPLICATION', 'SHUTDOWN', 'RESTART']
    assert [x.name for x in RunningFailureStrategies] == expected


def test_process_request_result():
    """ Test the ProcessRequestResult enumeration. """
    expected = ['IN_PROGRESS', 'SUCCESS', 'FAILED', 'TIMED_OUT']
    assert [x.name for x in ProcessRequestResult] == expected


def test_distribution_rules():
    """ Test the DistributionRules enumeration. """
    expected = ['ALL_INSTANCES', 'SINGLE_INSTANCE', 'SINGLE_NODE']
    assert [x.name for x in DistributionRules] == expected


def test_statistics_types():
    """ Test the StatisticsTypes enumeration. """
    expected = ['OFF', 'HOST', 'PROCESS', 'ALL']
    assert [x.name for x in StatisticsTypes] == expected


def test_synchronization_options():
    """ Test the SynchronizationOptions enumeration. """
    expected = ['STRICT', 'LIST', 'TIMEOUT', 'CORE', 'USER']
    assert [x.name for x in SynchronizationOptions] == expected


def test_publication_headers():
    """ Test the PublicationHeaders enumeration. """
    expected = ['TICK', 'PROCESS', 'PROCESS_ADDED', 'PROCESS_REMOVED',
                'PROCESS_DISABILITY', 'HOST_STATISTICS', 'PROCESS_STATISTICS', 'STATE']
    assert [x.name for x in PublicationHeaders] == expected


def test_request_headers():
    """ Test the RequestHeaders enumeration. """
    expected = ['CHECK_INSTANCE', 'START_PROCESS', 'STOP_PROCESS',
                'RESTART', 'SHUTDOWN', 'RESTART_SEQUENCE', 'RESTART_ALL', 'SHUTDOWN_ALL']
    assert [x.name for x in RequestHeaders] == expected


def test_notification_headers():
    """ Test the NotificationHeaders enumeration. """
    expected = ['AUTHORIZATION', 'STATE', 'ALL_INFO', 'DISCOVERY', 'INSTANCE_FAILURE']
    assert [x.name for x in NotificationHeaders] == expected


def test_event_headers():
    """ Test the RemoteCommEvents enumeration. """
    expected = ['SUPVISORS', 'INSTANCE', 'APPLICATION', 'PROCESS_EVENT', 'PROCESS_STATUS',
                'HOST_STATISTICS', 'PROCESS_STATISTICS']
    assert [x.name for x in EventHeaders] == expected


def test_invalid_transition():
    """ Test the exception InvalidTransition. """
    # test with unknown attributes
    with pytest.raises(InvalidTransition) as exc:
        raise InvalidTransition('invalid transition')
    assert 'invalid transition' == str(exc.value)


def test_parse_error():
    """ Test the exception ApplicationStatusParseError. """
    # test with unknown attributes
    with pytest.raises(ApplicationStatusParseError) as exc:
        raise ApplicationStatusParseError('parse error')
    assert 'parse error' == str(exc.value)


def test_supvisors_faults():
    """ Test the SupvisorsFaults enumeration. """
    expected = ['SUPVISORS_CONF_ERROR', 'BAD_SUPVISORS_STATE', 'NOT_MANAGED', 'DISABLED',
                'NOT_APPLICABLE', 'NOT_INSTALLED']
    assert [x.name for x in SupvisorsFaults] == expected
    assert [x.value for x in SupvisorsFaults] == list(range(100, 106))


def test_process_event():
    """ Test the ProcessEvent classes. """
    # attribute called 'name' cannot be mocked at creation
    process = Mock(config=Mock(), group=None)
    process.config.name = 'dummy_process'
    # test ProcessEvent creation
    event = ProcessEvent(process)
    assert isinstance(event, Event)
    assert event.process is process
    # test payload with no group
    assert event.payload() == 'processname:dummy_process groupname: '
    # test payload with no group
    process.group = Mock(config=Mock())
    process.group.config.name = 'dummy_group'
    assert event.payload() == 'processname:dummy_process groupname:dummy_group '
    # test ProcessAddedEvent creation
    event = ProcessAddedEvent(process)
    assert isinstance(event, ProcessEvent)
    assert event.payload() == 'processname:dummy_process groupname:dummy_group '
    # test ProcessRemovedEvent creation
    event = ProcessRemovedEvent(process)
    assert isinstance(event, ProcessEvent)
    assert event.payload() == 'processname:dummy_process groupname:dummy_group '


def test_supvisors_process_config():
    """ Test the SupvisorsProcessConfig class. """
    # first test the ProgramConfig class
    program_config = ProgramConfig('dummy_program', ProgramConfig)
    assert program_config.name == 'dummy_program'
    assert program_config.klass is ProgramConfig
    assert program_config.numprocs == 1
    assert program_config.group_config_info == {}
    assert not program_config.disabled
    # now test the SupvisorsProcessConfig class
    process_config = SupvisorsProcessConfig(program_config, 4, '-s test')
    assert process_config.program_config is program_config
    assert process_config.process_index == 4
    assert process_config.command_ref == '-s test'



# ======================================================================
# Copyright 2016 Julien LE CLEACH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ======================================================================

import ast
import re
from typing import Any, Dict, List, Optional, Sequence, Union

from supervisor.loggers import Logger
from supervisor.states import ProcessStates, RUNNING_STATES

from .process import ProcessStatus
from .ttypes import (ApplicationStates, DistributionRules, NameList, NameSet, Payload, StartingStrategies,
                     StartingFailureStrategies, RunningFailureStrategies, ApplicationStatusParseError)
from .utils import WILDCARD

# additional annotation types
ProcessList = List[ProcessStatus]
ProcessMap = Dict[str, ProcessStatus]  # {process_name: ProcessStatus}


class ApplicationRules:
    """ Definition of the rules for starting an application, iaw rules file.

    Attributes are:
        - managed: set to True when application rules are found from the rules file ;
        - distribution: the distribution rule of the application ;
        - identifiers: the Supervisors where the application can be started if not distributed (default: all) ;
        - hash_identifiers: when # rule is used, the application can be started on one of the Supervisors identified ;
        - start_sequence: defines the order of this application when starting all the applications
          in the DEPLOYMENT state (0 means: no automatic start) ;
        - stop_sequence: defines the order of this application when stopping all the applications
          (0 means: immediate stop) ;
        - starting_strategy: defines the strategy to apply when choosing the identifier where the process
          shall be started during the starting of the application ;
        - starting_failure_strategy: defines the strategy to apply when a required process cannot be started
          during the starting of the application ;
        - running_failure_strategy: defines the default strategy to apply when a required process crashes
          when the application is running ;
        - status_formula: the formula describing the application operational status ;
        - status_tree: the AST-parsed formula describing the application operational status.
    """

    # attributes
    managed: bool = False
    distribution: DistributionRules = DistributionRules.ALL_INSTANCES
    start_sequence: int = 0
    stop_sequence: int = -1
    starting_strategy: StartingStrategies = StartingStrategies.CONFIG
    starting_failure_strategy: StartingFailureStrategies = StartingFailureStrategies.ABORT
    running_failure_strategy: RunningFailureStrategies = RunningFailureStrategies.CONTINUE
    _status_formula: Optional[str] = None
    _status_tree: Optional[ast.Module] = None

    def __init__(self, supvisors: Any) -> None:
        """ Initialization of the attributes. """
        self.supvisors = supvisors
        # attributes with mutable default values
        self.identifiers: NameList = [WILDCARD]
        self.at_identifiers: NameList = []
        self.hash_identifiers: NameList = []

    @property
    def logger(self) -> Logger:
        """ Get the Supvisors logger. """
        return self.supvisors.logger

    @property
    def status_formula(self) -> str:
        """ Return the status formula if it has passed the AST parsing. """
        return self._status_formula

    @status_formula.setter
    def status_formula(self, formula: str):
        """ AST-parse the status formula. """
        # AST-parse the formula
        try:
            tree = ast.parse(formula)
        except SyntaxError:
            raise ApplicationStatusParseError('AST parse failure')
        # there must be only one element in the body
        if len(tree.body) != 1:
            raise ApplicationStatusParseError('unsupported AST expression')
        # store the expression
        self._status_formula = formula
        self._status_tree = tree

    @property
    def status_tree(self) -> Optional[ast.Expr]:
        """ Return the root expression of the AST-parsed status formula. """
        return self._status_tree.body[0].value if self._status_tree else None

    def check_stop_sequence(self, application_name: str) -> None:
        """ Check the stop_sequence value.
        If stop_sequence hasn't been set from the rules file, use the same value as start_sequence.

        :param application_name: the name of the application considered.
        :return: None
        """
        if self.stop_sequence < 0:
            self.logger.trace(f'ApplicationRules.check_stop_sequence: {application_name}'
                              f' - set stop_sequence to {self.start_sequence} ')
            self.stop_sequence = self.start_sequence

    def check_hash_identifiers(self, application_name: str) -> None:
        """ When a '#' is set in application rules, an association has to be done between the 'index' of the application
        and the index of the Supervisor in the applicable identifier list.
        Unlike programs, there is no unquestionable index that Supvisors could get because Supervisor does not support
        homogeneous applications. It thus has to be a convention.
        The chosen convention is that the application_name MUST match r'[-_]\\d+$'. The first index name is 1.

        hash_identifiers is expected to contain:
            - either ['*'] when all Supervisors are applicable
            - or any subset of these Supervisors.

        :param application_name: the name of the application considered.
        :return: None
        """
        error = True
        result = re.match(r'.*[-_](\d+)$', application_name)
        if not result:
            self.logger.error(f'ApplicationRules.check_hash_identifiers: {application_name} incompatible'
                              ' with the use of #')
        else:
            application_number = int(result.group(1)) - 1
            if application_number < 0:
                self.logger.error(f'ApplicationRules.check_hash_identifiers: index of {application_name} must be > 0')
            else:
                self.logger.trace(f'ApplicationRules.check_hash_identifiers: application_name={application_name}'
                                  f' application_number={application_number}')
                if '*' in self.hash_identifiers:
                    # all identifiers defined in the supvisors section of the supervisor configuration are applicable
                    ref_identifiers = list(self.supvisors.mapper.instances.keys())
                else:
                    # the subset of applicable identifiers is the hash_identifiers
                    ref_identifiers = self.hash_identifiers
                # if there are more application instances than possible identifiers, roll over
                index = application_number % len(ref_identifiers)
                self.identifiers = [ref_identifiers[index]]
                self.logger.debug(f'ApplicationRules.check_hash_identifiers: application_name={application_name}'
                                  f' identifiers={self.identifiers}')
                error = False
        if error:
            self.logger.debug(f'ApplicationRules.check_hash_identifiers: {application_name} start_sequence reset')
            self.start_sequence = 0

    def check_dependencies(self, application_name: str) -> None:
        """ Update rules after they have been read from the rules file.

        :param application_name: the name of the application considered.
        :return: None
        """
        self.check_stop_sequence(application_name)
        if self.hash_identifiers:
            self.check_hash_identifiers(application_name)

    def __str__(self) -> str:
        """ Get the application rules as string.

        :return: the printable application rules
        """
        return (f'managed={self.managed} distribution={self.distribution.name} identifiers={self.identifiers}'
                f' start_sequence={self.start_sequence} stop_sequence={self.stop_sequence}'
                f' starting_strategy={self.starting_strategy.name}'
                f' starting_failure_strategy={self.starting_failure_strategy.name}'
                f' running_failure_strategy={self.running_failure_strategy.name}'
                f' status_formula={self.status_formula}')

    # serialization
    def serial(self) -> Payload:
        """ Get a serializable form of the application rules.
        No information for unmanaged applications.

        :return: the application rules in a dictionary
        """
        if self.managed:
            payload = {'managed': True, 'distribution': self.distribution.name,
                       'identifiers': self.identifiers,
                       'start_sequence': self.start_sequence, 'stop_sequence': self.stop_sequence,
                       'starting_strategy': self.starting_strategy.name,
                       'starting_failure_strategy': self.starting_failure_strategy.name,
                       'running_failure_strategy': self.running_failure_strategy.name,
                       'status_formula': self.status_formula or ''}
            return payload
        return {'managed': False}


class HomogeneousGroup:
    """ Class used to store all the application processes that are configured using the same program definition.
    It has been introduced to manage the # rules optionally set to some processes of the application start sequence
    to distribute the processes over a set of Supvisors instances. """

    # information part
    program_name: str = ''
    # applicable process at/hash rules
    at_identifiers: Optional[NameList] = None
    hash_identifiers: Optional[NameList] = None

    def __init__(self, program_name: str, supvisors: Any) -> None:
        """ Initialization of the attributes.

        :param program_name: the name of the homogeneous group
        :param supvisors: the global Supvisors structure
        """
        self.supvisors = supvisors
        self.program_name = program_name
        # the list of processes belonging to the group
        self.processes: ProcessList = []

    @property
    def logger(self) -> Logger:
        """ Return the Supvisors logger. """
        return self.supvisors.logger

    def add_process(self, process: ProcessStatus) -> None:
        """ Add a process to the homogeneous group.
        process is expected to have the correct program_name (not checked here).

        :param process: the process belonging to the group
        :return: None
        """
        self.processes.append(process)
        # check at/hash rule consistence
        # NOTE: Based on the same program definition, is it possible to give multiple rules driven by patterns
        #       e.g. from program:prg with numprocs=30, patterns prg_0* / prog_1* / prg_2*
        if process.rules.at_identifiers:
            if self.at_identifiers is not None and self.at_identifiers != process.rules.at_identifiers:
                self.logger.error('HomogeneousGroup.add_process: inconsistent identifiers rules applied on'
                                  f' homogeneous group={self.program_name} - @={self.at_identifiers} vs '
                                  f' @={process.rules.at_identifiers}')
            self.at_identifiers = process.rules.at_identifiers
        if process.rules.hash_identifiers:
            if self.hash_identifiers is not None and self.hash_identifiers != process.rules.hash_identifiers:
                self.logger.error('HomogeneousGroup.add_process: inconsistent identifiers rules applied on'
                                  f' homogeneous group={self.program_name} - #={self.hash_identifiers} vs '
                                  f' #={process.rules.hash_identifiers}')
            self.hash_identifiers = process.rules.hash_identifiers
        if self.at_identifiers and self.hash_identifiers:
            self.logger.error('HomogeneousGroup.add_process: inconsistent identifiers rules applied on'
                              f' homogeneous group={self.program_name} - @={self.at_identifiers} and '
                              f' #={self.hash_identifiers}')
            self.hash_identifiers = None

    def remove_process(self, process: ProcessStatus) -> None:
        """ Remove a process from the homogeneous group.

        :param process: the process to remove from the group
        :return: None
        """
        self.processes.remove(process)

    def resolve_rules(self) -> None:
        """ When a '#' or a '@' is set in program rules, an association has to be made between the process_index
        of the process and the index of the Supvisors identifier in the applicable identifiers list.
        In this case, rules.hash_identifiers or rules.hash_identifiers is expected to contain:
            - either ['*'] when all Supervisors are applicable
            - or any subset of these Supervisors.

        The hash_identifiers were originally solved just after the Process rules were loaded because all Supvisors
        instances were known right from the start.

        The discovery mode changes things a bit. In this case, Supvisors instances are not known in advance.
        However, the list can only increase in time and the ordering is kept, even when considering auto-fencing.

        That's why the resolution logic can be kept quite identical as before. It just needs to be deferred,
        and eventually updated.

        NOTE: At resolution time, some of these processes may have already been started on the 'wrong' Supvisors
              instance using direct calls to Supervisor, or by setting their autostart option to True.
              The resolution logic in Supvisors does NOT consider this case because it is clearly a misuse.

        Updating the process rules has no effect on a running process. The update will effectively take place when the
        process will be started again.

        :return: None
        """
        if self.at_identifiers:
            self.assign_at_identifiers()
        if self.hash_identifiers:
            self.assign_hash_identifiers()

    def assign_at_identifiers(self) -> None:
        """ Assign identifiers wrt the '@' rule set on the homogeneous group.
        The assignment is strictly limited by the length of the identifiers list, without roll-over.
        If the number of candidate processes is greater than the candidate identifiers, the processes in excess
        cannot be started using Supvisors.

        :return: None
        """
        # get process list ordered by process_index
        process_list = sorted(self.processes, key=lambda x: x.process_index)
        # get unassigned processes
        unassigned_processes = [process for process in process_list
                                if process.rules.at_identifiers]
        if unassigned_processes:
            # get instances
            if WILDCARD in self.at_identifiers:
                # all identifiers defined in the supvisors section of the supervisor configuration file are applicable
                ref_identifiers = list(self.supvisors.mapper.instances.keys())
            else:
                # the subset of applicable identifiers is the second element of rule 'identifiers'
                # filter the unknown identifiers (or remaining aliases)
                ref_identifiers = self.supvisors.mapper.filter(self.at_identifiers)
            self.logger.debug(f'ProcessRules.assign_at_identifiers: program={self.program_name}'
                              f' ref_identifiers={ref_identifiers}')
            # the aim of at_identifiers is to distribute the processes over a list of Supvisors instances,
            #   without having 2 processes assigned to the same identifier
            # get assigned identifiers, assuming rules identifiers have size == 1
            assigned_identifiers = [process.rules.identifiers[0] for process in process_list
                                    if process.rules.identifiers]
            self.logger.debug(f'ProcessRules.assign_at_identifiers: program={self.program_name}'
                              f' assigned_identifiers={assigned_identifiers}')
            # deduce unassigned identifiers
            unassigned_identifiers = [identifier for identifier in ref_identifiers
                                      if identifier not in assigned_identifiers]
            self.logger.debug(f'ProcessRules.assign_at_identifiers: program={self.program_name}'
                              f' unassigned_identifiers={unassigned_identifiers}')
            # resolve what can be done
            for process, identifier in zip(unassigned_processes, unassigned_identifiers):
                process.rules.at_identifiers = []
                process.rules.identifiers = [identifier]
                self.logger.info(f'ProcessRules.assign_at_identifiers: namespec={process.namespec} assigned to'
                                 f' identifiers={process.rules.identifiers}')
            # WARN: even if not in discovery mode, do NOT remove the 'at' status from homogeneous group
            #       the assignment is final but the number of processes in a homogeneous group can change
            #       (through the update_numprocs XML-RPC)

    def assign_hash_identifiers(self) -> None:
        """ Assign identifiers wrt the '#' rule set on the homogeneous group.
        The assignment is NOT limited by the length of the identifiers list.
        If the number of candidate processes is greater than the candidate `identifiers`, the assignment is performed
        by rolling over the identifiers list.

        :return: None
        """
        # get process list ordered by process_index
        process_list = sorted(self.processes, key=lambda x: x.process_index)
        # get unassigned processes
        unassigned_processes = [process for process in process_list
                                if process.rules.hash_identifiers]
        if unassigned_processes:
            # get instances
            if WILDCARD in self.hash_identifiers:
                # all identifiers defined in the supvisors section of the supervisor configuration file are applicable
                ref_identifiers = list(self.supvisors.mapper.instances.keys())
            else:
                # the subset of applicable identifiers is the second element of rule 'identifiers'
                # filter the unknown identifiers (or remaining aliases)
                ref_identifiers = self.supvisors.mapper.filter(self.hash_identifiers)
            self.logger.debug(f'ProcessRules.assign_hash_identifiers: program={self.program_name}'
                              f' ref_identifiers={ref_identifiers}')
            # the aim of hash_identifiers is to distribute the processes over a list of Supvisors instances, so
            # unassigned processes will go to the least loaded Supvisors instance, wrt the homogeneous group considered
            process_count_per_instance = {identifier: [] for identifier in ref_identifiers}
            for process in process_list:
                if process.rules.identifiers:
                    # NOTE: assumed size == 1 and identifier in ref_identifiers
                    process_count_per_instance[process.rules.identifiers[0]].append(process)
            # re-arrange to keep ref_identifiers ordering
            process_count = [[len(process_count_per_instance[identifier]), identifier]
                             for identifier in ref_identifiers]
            self.logger.debug(f'ProcessRules.assign_hash_identifiers: program={self.program_name}'
                              f' process_count={process_count}')
            for process in unassigned_processes:
                # choose the Supvisors instance the least loaded (without sorting by identifier name)
                count, identifier = identifier_count = min(process_count, key=lambda x: x[0])
                process.rules.hash_identifiers = []
                process.rules.identifiers = [identifier]
                self.logger.info(f'ProcessRules.assign_hash_identifiers: namespec={process.namespec} assigned to'
                                 f' identifiers={process.rules.identifiers}')
                # increment identifier counter
                identifier_count[0] = count + 1
            # WARN: even if not in discovery mode, do NOT remove the 'at' status from homogeneous group
            #       the assignment is final but the number of processes in a homogeneous group can change
            #       (through the update_numprocs XML-RPC)


class ApplicationStatus:
    """ Class defining the status of an application in Supvisors.

    Attributes are:
        - application_name: the name of the application, corresponding to the Supervisor group name,
        - state: the state of the application in ApplicationStates,
        - major_failure: a status telling if a required process is stopped while the application is running,
        - minor_failure: a status telling if an optional process has crashed while the application is running,
        - processes: the map (key is process name) of the ProcessStatus belonging to the application,
        - process_groups: the ProcessStatus map, sorted as homogeneous groups,
        - rules: the ApplicationRules applicable to this application,
        - start_sequence: the sequencing to start the processes belonging to the application, as a dictionary.
        - stop_sequence: the sequencing to stop the processes belonging to the application, as a dictionary.

    For start and stop sequences, each entry is a list of processes having the same sequence order, used as key.
    """

    # types for annotations
    ApplicationSequence = Dict[int, ProcessList]
    PrintableApplicationSequence = Dict[int, Sequence[str]]
    HomogeneousGroupsMap = Dict[str, HomogeneousGroup]  # {program_name: HomogeneousGroup}

    # information part
    application_name: str = ''
    _state: ApplicationStates = ApplicationStates.STOPPED
    major_failure: bool = False
    minor_failure: bool = False
    # process part
    rules: Optional[ApplicationRules] = None

    def __init__(self, application_name: str, rules: ApplicationRules, supvisors: Any) -> None:
        """ Initialization of the attributes.

        :param application_name: the name of the application
        :param rules: the rules applicable to the application
        :param supvisors: the global Supvisors structure
        """
        self.supvisors = supvisors
        self.application_name = application_name
        self.rules = rules
        # attributes with mutable default values
        self.processes: ProcessMap = {}
        self.process_groups: ApplicationStatus.HomogeneousGroupsMap = {}
        self.start_sequence: ApplicationStatus.ApplicationSequence = {}
        self.stop_sequence: ApplicationStatus.ApplicationSequence = {}

    @property
    def logger(self) -> Logger:
        """ Get the Supvisors logger. """
        return self.supvisors.logger

    # status methods
    def running(self) -> bool:
        """ Return True if the application is running.

        :return: the running status of the application
        """
        return self.state in [ApplicationStates.STARTING, ApplicationStates.RUNNING]

    def stopped(self) -> bool:
        """ Return True if the application is stopped.

        :return: the stopped status of the application
        """
        return self.state == ApplicationStates.STOPPED

    def never_started(self) -> bool:
        """ Return True if the application has never been started.

        :return: True if the processes of the application have never been started
        """
        return all(info['state'] == ProcessStates.STOPPED and info['stop'] == 0
                   for proc in self.processes.values()
                   for info in proc.info_map.values())

    @property
    def state(self) -> ApplicationStates:
        """ Getter of state attribute.

         :return: the application state
         """
        return self._state

    @state.setter
    def state(self, new_state: ApplicationStates) -> None:
        """ Setter of state attribute.

        :param new_state: the new application state
        :return: None
        """
        if self._state != new_state:
            self._state = new_state
            self.logger.info(f'Application.state: {self.application_name} is {self.state.name}')

    def has_running_processes(self) -> bool:
        """ Check if one of the application processes is running.
        The application state may be STOPPED in this case if the running process is out of the starting sequence.

        :return: True if any of the application processes is running
        """
        return any(process.running() for process in self.processes.values())

    def get_operational_status(self) -> str:
        """ Get a description of the operational status of the application.

        :return: the operational status as string
        """
        if self.state == ApplicationStates.RUNNING:
            if self.major_failure:
                return 'Not Operational'
            if self.minor_failure:
                return 'Degraded'
            return 'Operational'
        return ''

    # serialization
    def serial(self) -> Payload:
        """ Get a serializable form of the application status.

        :return: the application status in a dictionary
        """
        return {'application_name': self.application_name, 'managed': self.rules.managed,
                'statecode': self.state.value, 'statename': self.state.name,
                'major_failure': self.major_failure, 'minor_failure': self.minor_failure}

    def __str__(self) -> str:
        """ Get the application status as string.

        :return: the printable application status
        """
        return (f'application_name={self.application_name} managed={self.rules.managed} state={self.state.name}'
                f' major_failure={self.major_failure} minor_failure={self.minor_failure}')

    # process methods
    def add_process(self, process: ProcessStatus) -> None:
        """ Add a new process to the process list.

        :param process: the process status to be added to the application
        :return: None
        """
        self.logger.trace(f'ApplicationStatus.add_process: {self.application_name}'
                          f' - adding process={process.process_name}')
        self.processes[process.process_name] = process
        # add to homogeneous groups
        group = self.process_groups.get(process.program_name)
        if not group:
            group = HomogeneousGroup(process.program_name, self.supvisors)
            self.process_groups[process.program_name] = group
        group.add_process(process)

    def remove_process(self, process_name: str) -> None:
        """ Remove a process from the process list.

        :param process_name: the process to be removed from the application
        :return: None
        """
        self.logger.trace(f'ApplicationStatus.remove_process: {self.application_name}'
                          f' - removing process={process_name}')
        process = self.processes.pop(process_name)
        # clear entry from homogeneous groups
        self.process_groups[process.program_name].remove_process(process)
        if not self.process_groups[process.program_name].processes:
            del self.process_groups[process.program_name]
        # re-evaluate sequences and status
        self.update_sequences()
        self.update()

    def resolve_rules(self):
        """ Call for hash identifiers resolution as the application is marked to be started in the Supvisors Starter.
        The resolution will be based on the current known list of Supvisors instances.
        """
        # get the programs involved in the starting sequence
        sequenced_programs = {process.program_name
                              for sub_seq in self.start_sequence.values()
                              for process in sub_seq}
        # set their rules identifiers
        for program_name in sequenced_programs:
            self.process_groups[program_name].resolve_rules()

    def possible_identifiers(self) -> NameList:
        """ Return the list of Supervisor identifiers where the application could be started, assuming that its
        processes cannot be distributed over multiple Supvisors instances.
        To achieve that, three conditions:
            - the Supervisor shall know all the application programs (real-time configuration) ;
            - the Supervisor identifier shall be allowed (explicitly or implicitly) in the rules file ;
            - the programs shall not be disabled.

        :return: the list of identifiers where the program could be started.
        """
        rules_identifiers = self.rules.identifiers
        if '*' in self.rules.identifiers:
            filtered_identifiers = list(self.supvisors.mapper.instances.keys())
        else:
            # filter the unknown identifiers (due to Supvisors discovery mode, any identifier may be lately known)
            filtered_identifiers = self.supvisors.mapper.filter(rules_identifiers)
        # get the identifiers of all application processes
        actual_identifiers: List[NameSet] = [{identifier
                                              for identifier, info in process.info_map.items()
                                              if not info['disabled']}
                                             for process in self.processes.values()]
        if actual_identifiers:
            # the common list is the intersection of all subsets
            actual_identifiers = actual_identifiers[0].intersection(*actual_identifiers)
        self.logger.debug(f'ApplicationStatus.possible_identifiers: application_name={self.application_name}'
                          f' actual_identifiers={actual_identifiers}')
        # intersect with rules
        return [identifier
                for identifier in filtered_identifiers
                if identifier in actual_identifiers]

    def possible_node_identifiers(self) -> NameList:
        """ Same principle as possible_identifiers, excepted that the possibilities are built from the nodes
        rather than from the strict identifiers.

        Note: Some elements in the returned list of identifiers may not fit to the possibilities got from the
              intermediate actual_identifiers, that is based on the real-time configurations.

        :return: the list of identifiers where the program could be started.
        """
        rules_identifiers = self.rules.identifiers
        if '*' in self.rules.identifiers:
            filtered_identifiers = list(self.supvisors.mapper.instances.keys())
        else:
            # filter the unknown identifiers (due to Supvisors discovery mode, any identifier may be lately known)
            filtered_identifiers = self.supvisors.mapper.filter(rules_identifiers)
        # get the identifiers of all application processes
        actual_identifiers: List[NameSet] = [{identifier
                                              for identifier, info in process.info_map.items()
                                              if not info['disabled']}
                                             for process in self.processes.values()]
        self.logger.trace(f'ApplicationStatus.possible_node_identifiers: application_name={self.application_name}'
                          f' actual_identifiers={actual_identifiers}')
        # test nodes individually
        all_node_identifiers = set()
        for node, identifiers_list in self.supvisors.mapper.nodes.items():
            # from all node identifiers, intersect with rules identifiers
            filtered_node_identifiers = {x for x in identifiers_list if x in filtered_identifiers}
            self.logger.trace(f'ApplicationStatus.possible_node_identifiers: application_name={self.application_name}'
                              f' node={node} filtered_node_identifiers={filtered_node_identifiers}')
            # check that every process has a solution on the node
            config_identifiers = set()
            for process_identifiers in actual_identifiers:
                process_node_identifiers = filtered_node_identifiers & set(process_identifiers)
                if not process_node_identifiers:
                    self.logger.debug('ApplicationStatus.possible_node_identifiers:'
                                      f' application_name={self.application_name} has no solution on node={node}')
                    break
                # add remaining identifiers to the intersections union
                config_identifiers.update(process_node_identifiers)
            else:
                self.logger.debug('ApplicationStatus.possible_node_identifiers:'
                                  f' application_name={self.application_name} node={node}'
                                  f' solution={config_identifiers}')
                # add the node solution to the application solutions
                all_node_identifiers.update(config_identifiers)
        self.logger.debug(f'ApplicationStatus.possible_node_identifiers: application_name={self.application_name}'
                          f' all_node_identifiers={all_node_identifiers}')
        # return the node identifiers keeping the ordering defined in rules
        return [identifier
                for identifier in filtered_identifiers
                if identifier in all_node_identifiers]

    def get_instance_processes(self, identifier: str) -> ProcessList:
        """ Return the list of application processes configured in the Supervisor instance.

        :param identifier: the Supervisor instance where the processes qre configured
        :return: the list of application processes in the Supervisor instance
        """
        return [process for process in self.processes.values()
                if identifier in process.info_map]

    # sequencing methods
    @staticmethod
    def printable_sequence(application_sequence: ApplicationSequence) -> PrintableApplicationSequence:
        """ Get printable application sequence for log traces.
        Only the name of the process is kept.

        :return: the simplified sequence
        """
        return {seq: [proc.process_name for proc in proc_list]
                for seq, proc_list in application_sequence.items()}

    def update_sequences(self) -> None:
        """ Evaluate the sequencing of the starting / stopping application from its list of processes.
        This makes sense only for managed applications.

        :return: None
        """
        self.start_sequence.clear()
        self.stop_sequence.clear()
        # consider only managed applications for start sequence
        if self.rules.managed:
            # fill ordering iaw process rules
            for process in self.processes.values():
                self.start_sequence.setdefault(process.rules.start_sequence, []).append(process)
            self.logger.debug(f'ApplicationStatus.update_sequences: application_name={self.application_name}'
                              f' start_sequence={self.printable_sequence(self.start_sequence)}')
        else:
            self.logger.info(f'ApplicationStatus.update_sequences: application_name={self.application_name}'
                             ' is not managed so start sequence is undefined')
        # stop sequence is applicable to all applications
        for process in self.processes.values():
            # fill ordering iaw process rules
            self.stop_sequence.setdefault(process.rules.stop_sequence, []).append(process)
        self.logger.debug(f'ApplicationStatus.update_sequences: application_name={self.application_name}'
                          f' stop_sequence={self.printable_sequence(self.stop_sequence)}')

    def get_start_sequenced_processes(self) -> ProcessList:
        """ Return the processes included in the application start sequence.
        The sequence 0 is removed as it corresponds to processes that are not meant to be auto-started.

        :return: the processes included in the application start sequence.
        """
        return [process for seq, sub_seq in self.start_sequence.items()
                for process in sub_seq if seq > 0]

    def get_start_sequence_expected_load(self) -> int:
        """ Return the sum of the expected loading of the processes in the starting sequence.
        This is used only in the event where the application is not distributed and the whole application loading
        has to be checked on a single Supervisor.

        :return: the expected loading of the application.
        """
        return sum(process.rules.expected_load for process in self.get_start_sequenced_processes())

    # global status methods
    def update(self) -> None:
        """ Update the state and the operational status of the application.

        :return: None
        """
        # get the processes from the starting sequence
        sequenced_processes: ProcessMap = {process.process_name: process
                                           for sub_seq in self.start_sequence.values()
                                           for process in sub_seq}
        # update the application state
        self.state = self.update_state()
        # reset failures
        self.major_failure, self.minor_failure = False, False
        # update the application operational status
        if self.rules.status_tree:
            self.update_status_formula(sequenced_processes)
        else:
            self.update_status_required(sequenced_processes)
        self.logger.debug(f'ApplicationStatus.update: application_name={self.application_name}'
                          f' major_failure={self.major_failure} minor_failure={self.minor_failure}')

    def update_state(self) -> ApplicationStates:
        """ Update the state of the application iaw the state of its processes.
        This is a displayed status, and thus relies on the displayed process state, considering the forced state.

        :return: None
        """
        starting, running, stopping = False, False, False
        # evaluate the application state from the state of its processes
        for process in self.processes.values():
            self.logger.trace(f'ApplicationStatus.update_state: application_name={self.application_name}'
                              f' process={process.namespec} displayed_state={process.displayed_state}')
            if process.displayed_state == ProcessStates.RUNNING:
                running = True
            elif process.displayed_state in [ProcessStates.STARTING, ProcessStates.BACKOFF]:
                starting = True
            elif process.displayed_state == ProcessStates.STOPPING:
                stopping = True
        self.logger.trace(f'ApplicationStatus.update_state: application_name={self.application_name}'
                          f' - starting={starting} running={running} stopping={stopping}')
        # apply rules for state
        if stopping:
            # if at least one process is STOPPING, let's consider that application is stopping
            # here priority is given to STOPPING over STARTING
            return ApplicationStates.STOPPING
        if starting:
            # if at least one process is STARTING, let's consider that application is starting
            return ApplicationStates.STARTING
        if running:
            # all processes in the sequence are RUNNING, so application is RUNNING
            return ApplicationStates.RUNNING
        # all processes in the sequence are STOPPED, so application is STOPPED
        return ApplicationStates.STOPPED

    def update_status_required(self, sequenced_processes: ProcessMap) -> None:
        """ Update the operational status of the application iaw its processes' state.
        Any required process in failure implies a major failure.
        Any non-required process in failure implies a minor failure if part of the starting sequence.

        NOTE: This is the legacy method to evaluate the application status.
              The formula method is now preferred.

        :return: None
        """
        possible_major_failure = False
        for process in self.processes.values():
            self.logger.trace(f'ApplicationStatus.update_status_required: application_name={self.application_name}'
                              f' process={process.namespec} state={process.displayed_state_string()}'
                              f' required={process.rules.required} exit_expected={process.expected_exit}')
            if (process.displayed_state in [ProcessStates.FATAL, ProcessStates.UNKNOWN]
                    or (process.displayed_state == ProcessStates.EXITED and not process.expected_exit)):
                # a major/minor failure is raised in these states depending on the required option
                if process.rules.required:
                    self.major_failure = True
                elif process.process_name in sequenced_processes:
                    self.minor_failure = True
            elif process.displayed_state == ProcessStates.STOPPED:
                # possible major failure raised if process STOPPED
                # consideration will depend on the global application state
                if process.rules.required:
                    possible_major_failure = True
        self.logger.trace(f'ApplicationStatus.update_status_required: application_name={self.application_name}'
                          f' major_failure={self.major_failure} minor_failure={self.minor_failure}'
                          f' possible_major_failure={possible_major_failure}')
        # possible major failure is confirmed if the application is not stopped
        if self.state != ApplicationStates.STOPPED:
            self.major_failure |= possible_major_failure
        # reset the minor failure if the major failure is set
        if self.major_failure:
            self.minor_failure = False

    def evaluate(self, node: ast.Expr) -> Union[bool, List[bool]]:
        """ Resolution of the AST expression.

        All tree nodes are expected to be:
            * a boolean operator (or, and) ;
            * a unary operator (not) ;
            * a function (any, all).

        All tree leaves are expected to be string corresponding to:
            * a process name ;
            * a pattern matching at least one process name.

        When multiple process names are found from a pattern, it is expected that a function evaluates them.

        :param node: The current AST node.
        :return: Either a boolean value or a list a boolean values intended to be used by an upper Call.
        """
        # handle string leaves (ast.Str is deprecated from Python < 3.8)
        leaf: Optional[str] = None
        if type(node) is ast.Str:
            leaf = node.s
        elif type(node) is ast.Constant and type(node.value) is str:
            leaf = node.value
        if leaf is not None:
            if leaf in self.processes:
                return self._get_process_status(leaf)
            matches = self._get_matches(leaf)
            if len(matches) == 1:
                return self._get_process_status(matches[0])
            elif len(matches) >= 1:
                return [self._get_process_status(x) for x in matches]
            raise ApplicationStatusParseError(f'no match for expression={node.s}')
        # handle any/all functions
        if type(node) is ast.Call:
            if node.func.id not in ['all', 'any']:
                raise ApplicationStatusParseError(f'unsupported function={node.func.id}')
            args_eval = self.evaluate(node.args[0])
            if type(args_eval) is bool:
                args_eval = [args_eval]
            return eval(f'{node.func.id}({args_eval})')
        # handle and/or operators
        if type(node) is ast.BoolOp:
            args_eval = [self.evaluate(x) for x in node.values]
            if any(type(x) is not bool for x in args_eval):
                raise ApplicationStatusParseError(f'cannot apply BoolOp on unresolved expression')
            if type(node.op) is ast.And:
                return all(args_eval)
            if type(node.op) is ast.Or:
                return any(args_eval)
        # handle not operator
        if type(node) is ast.UnaryOp:
            if type(node.op) is ast.Not:
                args_eval = self.evaluate(node.operand)
                if type(args_eval) is not bool:
                    raise ApplicationStatusParseError(f'cannot apply UnaryOp on unresolved expression')
                return not args_eval
            raise ApplicationStatusParseError(f'unsupported UnaryOp={type(node.op).__name__}')
        raise ApplicationStatusParseError(f'unsupported Expr={type(node).__name__}')

    def update_status_formula(self, sequenced_processes: ProcessMap) -> None:
        """ Update the operational status of the application iaw its processes' state.
        The major failure is set according to the status formula.
        If no major failure, any process in failure implies a minor failure if part of the starting sequence.

        :return: None
        """
        # evaluate the status formula against the current processes' state
        try:
            result = self.evaluate(self.rules.status_tree)
            if type(result) is not bool:
                raise ApplicationStatusParseError('status formula cannot be resolved')
        except ApplicationStatusParseError as exc:
            self.logger.error(f'ApplicationStatus.update_status_formula: application_name={self.application_name}'
                              f' - {exc}')
            self.major_failure = True
        else:
            self.major_failure = not result
        # check for a minor failure in start sequence processes
        if not self.major_failure:
            for process in sequenced_processes.values():
                self.logger.trace(f'ApplicationStatus.update_status_formula: application_name={self.application_name}'
                                  f' process={process.namespec} state={process.displayed_state_string()}'
                                  f' exit_expected={process.expected_exit}')
                if (process.displayed_state in [ProcessStates.FATAL, ProcessStates.UNKNOWN]
                        or (process.displayed_state == ProcessStates.EXITED and not process.expected_exit)):
                    self.minor_failure = True

    def _get_process_status(self, process_name: str) -> bool:
        """ Return False if the process is in a stopped (displayed) state.
        Exception is made for EXITED expected, which is considered as normal. """
        process = self.processes[process_name]
        return ((process.displayed_state in RUNNING_STATES)
                or (process.displayed_state == ProcessStates.EXITED and process.expected_exit))

    def _get_matches(self, pattern_name: str) -> List[str]:
        """ Return the process names matching the pattern. """
        results = []
        pattern = re.compile(r'^%s$' % pattern_name)
        for name in self.processes.keys():
            if pattern.match(name):
                results.append(name)
        return results
